/*
    FEARLUS model0-7: FearlusArguments.m
    Copyright (C) 1999-2003  Macaulay Institute

    This file is part of FEARLUS model0-7, an agent-based model of land use
    change.

    FEARLUS model0-7 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model0-7 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*


Implementation of the FearlusArguments class -- a class to parse any command
line arguments to be implemented by the model. Pretty much all copied from the
Arguments page in the Swarm documentation (Example objectbase/Arguments/1)

model0-5-3-T: Changed to be a single-instance class to allow other classes to
access FEARLUS-specific information about the arguments through the class
rather than through the Swarm-provided global variable 'arguments'. This is
required to stop the compiler complaining that the arguments variable, which
follows the Arguments protocol, doesn't implement (necessarily) any of the
FEARLUS-specific methods for argument information.

*/

#import "FearlusArguments.h"
#import <sys/time.h>
#define DEBUG_PARCELS 0x1
#define DEBUG_USES 0x2
#define DEBUG_MANAGERS 0x4
#define DEBUG_ALLOCATOR 0x8
#define DEBUG_ENVIRONMENT 0x10
#define DEBUG_OBSERVER 0x20
#define DEBUG_MODEL 0x40
#define DEBUG_CORE_METHODS 0x80
#define DEBUG_INNER_CORE_METHODS 0x100	// Neighbourhood & sorting
#define DEBUG_NEIGHBOURHOOD 0x200	// Specific neighbourhood messages
#define DEBUG_BITSTRINGS 0x400		// Bitstring specific messages

static FearlusArguments *squabbles;
static BOOL instantiated = NO;
static const char *seedArg;
static BOOL seedSpecified;
static const char *postInitSeedArg;
static BOOL postInitSeedSpecified;
static const char *modelFile;
static const char *historyFile;
static BOOL historyFileSpecified;
static const char *landParcelFile;
static BOOL landParcelFileSpecified;
static const char *landUseFile;
static BOOL landUseFileSpecified;
static int debugFlags;
static BOOL dosMode;
static const char *reporterFile;
static BOOL reporterFileSpecified;
static const char *reporterCFFile;
static BOOL reporterCFFileSpecified;
static BOOL appendReport;
static const char *observerFile;
static BOOL observerFileSpecified;

static BOOL seed1set = NO, seed2set = NO;
static unsigned seed1, seed2;

@implementation FearlusArguments

/*

createBegin:

This method is copied straight out of the Arguments class page in the Swarm
documentation, with minor changes as appropriate.

*/

+createBegin: (id)aZone {
  static struct argp_option options[] = {
    {"withseed", 'X', "SEED", 0, "Specify the seed. [0-9]+, TIME or DEFAULT",
     99},
    {"postinitseed", 'Z', "POSTINITSEED", 0, "Specify a separate seed to use "
     "after initialisation: [0-9]+ or TIME", 99},
    {"parameters", 'p', "PARAMETERFILE", 0,
     "File for the model parameters to be loaded from", 99},
    {"observers", 'o', "OBSERVERFILE", 0,
     "File for the observer settings to be loaded from", 99},
    {"history", 'h', "HISTORYFILE", 0,
     "File for the history to be saved to", 99},
    {"dos", 'd', NULL, 0,
     "DOS mode history file output -- end of line = CR+LF not just LF", 99},
    {"parcels", 'P', "LANDPARCELFILE", 0,
     "File to load the land parcel bit strings from", 99},
    {"uses", 'U', "LANDUSEFILE", 0,
     "File to load the land use bit strings from", 99},
    {"report", 'r', "REPORTOUTPUTFILE", 0, "File to save the report to", 99},
    {"repconfig", 'R', "REPORTCONFIGFILE", 0,
     "Reporter configuration file", 99},
    {"append-report", 'a', NULL, 0, "If report file exists, then append to it",
     99},
    {"debug", 'D', "DEBUGFLAGS", 0,
     "Debug flags [P]arcel [U]ses [M]anagers [A]llocator [E]nvironment "
     "[o]bserver [m]odel [c]ore methods [i]nner core methods", 99},
    { 0 }
  };

  if(instantiated) {
    fprintf(stderr, "WARNING: createBegin of FearlusArguments called again\n");
  }
  squabbles = [super createBegin: aZone];
  instantiated = YES;

  [squabbles addOptions: options];
  seedSpecified = NO;
  postInitSeedSpecified = NO;
  modelFile = "fearlus.model";
  historyFile = "fearlus-history.txt";
  historyFileSpecified = NO;
  landParcelFile = "fearlus.lp";
  landParcelFileSpecified = NO;
  landUseFile = "fearlus.lu";
  landUseFileSpecified = NO;
  debugFlags = 0;
  dosMode = NO;
  reporterFileSpecified = NO;
  reporterFile = "fearlus-report.txt";
  reporterCFFileSpecified = NO;
  reporterCFFile = "fearlus-report.cf";
  appendReport = NO;
  observerFileSpecified = NO;
  observerFile = "fearlus.obs";
  return squabbles;
}

/*

parseKey:arg:

Parse the argument for the given key. Switch used to enable easy expansion.

*/

-(int)parseKey: (int)key arg: (const char *)arg {
  switch(key) {
  case 'X':
    seedArg = arg;
    seedSpecified = YES;
    return 0;
  case 'Z':
    postInitSeedArg = arg;
    postInitSeedSpecified = YES;
    return 0;
  case 'h':
    historyFile = arg;
    historyFileSpecified = YES;
    return 0;
  case 'o':
    observerFile = arg;
    observerFileSpecified = YES;
    return 0;
  case 'p':
    modelFile = arg;
    return 0;
  case 'P':
    landParcelFile = arg;
    landParcelFileSpecified = YES;
    return 0;
  case 'U':
    landUseFile = arg;
    landUseFileSpecified = YES;
    return 0;
  case 'r':
    reporterFile = arg;
    reporterFileSpecified = YES;
    return 0;
  case 'R':
    reporterCFFile = arg;
    reporterCFFileSpecified = YES;
    return 0;
  case 'a':
    appendReport = YES;
    return 0;
  case 'D':
    [self parseDebugFlags: arg];
    return 0;
  case 'd':
    dosMode = YES;
    return 0;
  default:
    return [super parseKey: key arg: arg];
  }
}

/*

parseDebugFlags:

Parse the debug flags given in the argument.

*/

-(void)parseDebugFlags: (const char *)arg {
  char *p;

  p = (char *)arg;
  while(*p != '\0') {
    switch(*p) {
    case 'P':
      debugFlags |= DEBUG_PARCELS;
      break;
    case 'U':
      debugFlags |= DEBUG_USES;
      break;
    case 'M':
      debugFlags |= DEBUG_MANAGERS;
      break;
    case 'A':
      debugFlags |= DEBUG_ALLOCATOR;
      break;
    case 'E':
      debugFlags |= DEBUG_ENVIRONMENT;
      break;
    case 'o':
      debugFlags |= DEBUG_OBSERVER;
      break;
    case 'm':
      debugFlags |= DEBUG_MODEL;
      break;
    case 'c':
      debugFlags |= DEBUG_CORE_METHODS;
      break;
    case 'i':
      debugFlags |= DEBUG_INNER_CORE_METHODS;
      break;
    case 'n':
      debugFlags |= DEBUG_NEIGHBOURHOOD;
      break;
    case 'b':
      debugFlags |= DEBUG_BITSTRINGS;
      break;
    default:
      fprintf(stderr, "Debug flag not recognised: %c\n", *p);
    }
    p++;
  }
}

/*

seedHasBeenSpecified

Return a boolean value specifying whether or not a seed has been specified on
the command line argument. Returns true for [0-9]+ or TIME values, false
otherwise.

*/

+(BOOL)seedHasBeenSpecified {
  if(seedSpecified) {
    if(!strcmp(seedArg, "TIME")) {
      return YES;
    }
    else if(!strcmp(seedArg, "DEFAULT")) {
      return NO;
    }
    else {
      const char *p = seedArg;

      while(*p) {
	if(!isdigit((int)*p)) {
	  fprintf(stderr, "Invalid value for seed argument: %s -- using "
		  "default value\n", seedArg);
	  return NO;
	}
	p++;
      }
      return YES;
    }
  }
  else {
    return NO;
  }
}

/*

getSeedArg

Return the seed argument value -- this is either the number specified, or
if the seedArg is "TIME", then use the current internal clock time as a seed.

*/

+(unsigned)getSeedArg {
  unsigned seed;

  if(!seedSpecified || !strcmp(seedArg, "DEFAULT")) {
    fprintf(stderr, "Cannot get the seed value when it isn't specified!\n");
    return 0;
  }
  if(!strcmp(seedArg, "TIME")) {
    time_t now = time(0);

    seed = (unsigned)now;
    if(seed2set && seed == seed2) {
      seed++;
    }
    seed1 = seed;
    seed1set = YES;
  }
  else {
    seed = (unsigned)strtoul(seedArg, NULL, 10);
  }
  return seed;
}

/*

postInitseedHasBeenSpecified

Return a boolean value specifying whether or not a separate seed to use after
initialisation has been given on the command line. Bomb out if an invalid
argument has been given.

*/

+(BOOL)postInitSeedHasBeenSpecified {
  if(postInitSeedSpecified) {
    if(!strcmp(postInitSeedArg, "TIME")) {
      return YES;
    }
    else {
      const char *p = postInitSeedArg;

      while(*p) {
	if(!isdigit((int)*p)) {
	  fprintf(stderr, "Invalid value for seed argument: %s\n",
		  postInitSeedArg);
	  abort();
	}
	p++;
      }
      return YES;
    }
  }
  else {
    return NO;
  }
}

/*

getPostInitSeedArg

Return the separate seed argument value -- this is either the number
specified, or if the seedArg is "TIME", then a value computed from the
result of the current clock time -- note that this needs to be
different from that used by getSeedArg, so we multiply by microseconds.

*/

+(unsigned)getPostInitSeedArg {
  unsigned seed;

  if(!postInitSeedSpecified) {
    fprintf(stderr, "Cannot get the seed value when it isn't specified!\n");
    abort();
  }
  if(!strcmp(postInitSeedArg, "TIME")) {
    struct timeval now;

    if(gettimeofday(&now, NULL) < 0) {
      perror("Setting seed: gettimeofday");
      abort();
    }
    
    now.tv_usec = (now.tv_usec == 0 ? now.tv_usec + 1 : now.tv_usec);
    seed = (unsigned)now.tv_sec * (unsigned)now.tv_usec;
    if(seed1set && seed == seed1) {
      seed++;
    }
    seed2 = seed;
    seed2set = YES;
  }
  else {
    seed = (unsigned)strtoul(postInitSeedArg, NULL, 10);
  }
  return seed;
}

/*

getModelFile

Return the model file to check for model parameters (fearlus.model by default)

*/

+(const char *)getModelFile {
  return modelFile;
}

/*

observerFileHasBeenSpecified

Return a value indicating whether or not the observer file was specified on
the command line. This is used to determine whether or not to load observer
settings.

*/

+(BOOL)observerFileHasBeenSpecified {
  return observerFileSpecified;
}

/*

getObserverFile

Return the name passed in as the observer file to use

*/

+(const char *)getObserverFile {
  return observerFile;
}

/*

historyFileHasBeenSpecified

Return a value indicating whether or not the history file was specified on the
command line. This is used to determine whether or not to save the history.

*/

+(BOOL)historyFileHasBeenSpecified {
  return historyFileSpecified;
}

/*

getHistoryFile

Return the history file specified on the command line. If none specified, it
will be the default value.

*/

+(const char *)getHistoryFile {
  return historyFile;
}

/*

landParcelFileHasBeenSpecified

Return a value indicating whether or not a land parcel file was specified on
the command line. This is used to determine whether or not to load the
land parcel bit strings as opposed to creating them at random. NB No clumping
is done when the land parcel bit strings are loaded.

*/

+(BOOL)landParcelFileHasBeenSpecified {
  return landParcelFileSpecified;
}

/*

getLandParcelFile

Return the land parcel file. If none specified, return the default one.

*/

+(const char *)getLandParcelFile {
  return landParcelFile;
}

/*

landUseFileHasBeenSpecified

Return a value indicating whether or not a land use file was specified on the
command line. This is used to determine whether or not to load the land use
bit strings as opposed to creating them at random.

*/

+(BOOL)landUseFileHasBeenSpecified {
  return landUseFileSpecified;
}

/*

getLandUseFile

Return the land use file. If none specified, return the default one.

*/

+(const char *)getLandUseFile {
  return landUseFile;
}

/*

reporterFileHasBeenSpecified

Return a Boolean value indicating whether or not a reporter file has been
specified on the command line.

*/

+(BOOL)reporterFileHasBeenSpecified {
  return reporterFileSpecified;
}

/*

getReporterFile

Return the reporter file. If none has been specified on the command line,
then return the default.

*/

+(const char *)getReporterFile {
  return reporterFile;
}

/*

reporterCFFileHasBeenSpecified

Return a Boolean value indicating whether or not a reporter configuration file
has been specified on the command line.

*/

+(BOOL)reporterCFFileHasBeenSpecified {
  return reporterCFFileSpecified;
}

/*

getReporterCFFile

Return the reporter configuration file. If none has been specified on the
command line, then return the default.

*/

+(const char *)getReporterCFFile {
  return reporterCFFile;
}

/*

useReporter

Return a Boolean value indicating whether or not a reporter has been requested.
This will be the case if either the reporter output or configuration files or
both have been specified. (The default output is stdout.)

*/

+(BOOL)useReporter {
  if(reporterCFFileSpecified || reporterFileSpecified) {
    return YES;
  }
  else {
    return NO;
  }
}

/*

appendReportFile

Return a Boolean value indicating whether or not the report file is to be
appended to if specified.

*/

+(BOOL)appendReportFile {
  return appendReport;
}

/*

debugParcels

Return a flag indicating whether debug messages are required for land parcels

*/

+(BOOL)debugParcels {
  return ((debugFlags & DEBUG_PARCELS) ? YES : NO);
}

/*

debugUses

Return a flag indicating whether debug messages are required for land uses

*/

+(BOOL)debugUses {
  return ((debugFlags & DEBUG_USES) ? YES : NO);
}

/*

debugManagers

Return a flag indicating whether debug messages are required for land managers

*/

+(BOOL)debugManagers {
  return ((debugFlags & DEBUG_MANAGERS) ? YES : NO);
}

/*

debugAllocator

Return a flag indicating whether debug messages are required for the land
allocator

*/

+(BOOL)debugAllocator {
  return ((debugFlags & DEBUG_ALLOCATOR) ? YES : NO);
}

/*

debugEnvironment

Return a flag indicating whether debug messages are required for the 
environment

*/

+(BOOL)debugEnvironment {
  return ((debugFlags & DEBUG_ENVIRONMENT) ? YES : NO);
}

/*

debugObserver

Return a flag indicating whether debug messages are required for the observer
swarm

*/

+(BOOL)debugObserver {
  return ((debugFlags & DEBUG_OBSERVER) ? YES : NO);
}

/*

debugModel

Return a flag indicating whether debug messages are required for the model
swarm

*/

+(BOOL)debugModel {
  return ((debugFlags & DEBUG_MODEL) ? YES : NO);
}

/*

debugCoreMethods

Return a flag indicating whether debug messages are required for certain core
methods

*/

+(BOOL)debugCoreMethods {
  return ((debugFlags & DEBUG_CORE_METHODS) ? YES : NO);
}

/*

debugInnerCoreMethods

Return a flag indicating whether debug messages are required for other, more
internal core methods

*/

+(BOOL)debugInnerCoreMethods {
  return ((debugFlags & DEBUG_INNER_CORE_METHODS) ? YES : NO);
}

/*

debugNeighbourhood

Return a flag indicating whether debug messages are required for looking at
neighbourhoods

*/

+(BOOL)debugNeighbourhood {
  return ((debugFlags & DEBUG_NEIGHBOURHOOD) ? YES : NO);
}

/*

debugBitstrings

Return a flag indicating whether debug messages are required for looking at
bitstrings

*/

+(BOOL)debugBitstrings {
  return ((debugFlags & DEBUG_BITSTRINGS) ? YES : NO);
}

/*

nl

Return a string of characters to put a new line in a history file. In DOS mode,
that string will be "\r\n", otherwise the default is "\n"

*/

+(const char *)nl {
  return ((dosMode) ? "\r\n" : "\n");
}

@end
