/*
    FEARLUS model0-7: SubPopulation.h
    Copyright (C) 1999-2003  Macaulay Institute

    This file is part of FEARLUS model0-7, an agent-based model of land use
    change.

    FEARLUS model0-7 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model0-7 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*


Interface for the SubPopulation object -- actually a set of parameters
describing different decision-making algorithms. Each land manager
will belong to a given sub-population, using that algorithm to make
their land use decision. Each SubPopulation specifies a range of
values for each of the habit-imitation threshold, the neighbourhood
weighting, and the imitative strategy probability.

In addition, there should be a GUI for entering all these different
values, and for displaying land managers (i.e. each subPopulation
should have a colour). The land allocator, when creating land
managers, should select a SubPopulation at random (according to a
given probability), and hand that SubPopulation to the land manager,
who will use it to get their decision-making algorithm.

For now, the GUI is just a probe map with a message probe for getting
up the strategy selector GUI. The probe map must get displayed when
another method is called, which is achieved when pressing a button on
the parameter object.

NOTE: A legacy from earlier versions is the use of "habit threshold"
or "habit imitation threshold" to refer to what we now call
"aspiration threshold". Various method and variable names referring to
"habit" should be interpreted as referring to "aspiration".

*/

#import <objectbase/SwarmObject.h>
#import <collections/List.h>
#import <simtools.h>
#import <simtoolsgui.h>
#import <errno.h>
#import <string.h>
#import <random.h>
#import "Strategy.h"

@class LandManager, Parameter, StrategySelector, Interval;

@interface SubPopulation: SwarmObject {
  double probability;	// Probability that this SubPopulation will be used to
                        // create a land manager.
  int colour;		// Index in a colourmap to use for this SubPopulation.
  StrategySelector *strategies;		// Strategy selector for this sub-pop.
  unsigned nRemovals;		// Number of removals of land managers
  char *landMgrClass;		// Land manager class to use
  double neighbourWeightMax;	// Maximum land manager neighbourhood weight
  double neighbourWeightMin;	// Minimum land manager neighbourhood weight
  double neighbourWeightMean;	// Mean ditto if using normal dist
  double neighbourWeightVar;	// Variance of ditto
  char *neighbourWeightDist;	// Which dist to use (normal or uniform)
  double imitateProbMax;	// Maximum probability of imitative strategy
  double imitateProbMin;	// Minimum probability of imitative strategy
  double imitateProbMean;	// Mean ditto if using normal dist
  double imitateProbVar;	// Variance of ditto
  char *imitateProbDist;	// Which dist to use (normal or uniform)
  double habitImitateThresholdMax;
				// Maximum aspiration threshold
  double habitImitateThresholdMin;
				// Minimum aspiration threshold
  double habitImitateThresholdMean;
				// Mean ditto if using normal dist
  double habitImitateThresholdVar;
				// Variance of ditto
  char *habitImitateThresholdDist;
				// Which dist to use (normal or uniform)
  unsigned memorySizeMin;	// Minimum size of memory
  unsigned memorySizeMax;	// Maximum size of memory
  char *strategySelectorFile;	// File to load the strategy selector data from
  unsigned pin;
  char *label;
  id <List> managerList;
  BOOL selectorCreated;
  id <NormalDist> normalDist;	// Normal distribution
}

+create: aZone;
-(unsigned)getPIN;
-(char *)getLabel;		// Must be created and allocated at create time
-(void)setColour: (int)col;
-(int)getColour;
-(unsigned)getNRemovals;
-resetRemovals;			// Used by model schedule
-(Class)getLandManagerClass;
-(Interval *)getANeighbourWeight;
-(double)getNeighbourWeightMin;	// Used by observer
-(double)getNeighbourWeightMax;	// Used by observer
-(char *)getNeighbourWeightDist;
-(double)getNeighbourWeightMean;
-(double)getNeighbourWeightVar;
-(Interval *)getAnImitateProb;
-(double)getImitateProbMin;	// Used by observer
-(double)getImitateProbMax;	// Used by observer
-(char *)getImitateProbDist;
-(double)getImitateProbMean;
-(double)getImitateProbVar;
-(Interval *)getAHabitImitateThreshold; /* aspiration */
-(double)getHabitThresholdMin;	// Used by observer // aspiration
-(double)getHabitThresholdMax;	// Used by observer // aspiration
-(char *)getHabitThresholdDist; // aspiration
-(double)getHabitThresholdMean; // aspiration
-(double)getHabitThresholdVar;  // aspiration
-(unsigned)getAMemorySize;
-(unsigned)getMaxMemorySize;
-(void)addLandManager: (LandManager *)lm;
-(void)removeLandManager: (LandManager *)lm;
-(id <List>)getLandManagers;
-(id <Strategy>)getAnAboveStrategyForLandManager: (LandManager *)lm
				   andParameters: (Parameter *)p;
-(id <NonImitativeStrategy>)
     getABelowNonImitativeStrategyForLandManager: (LandManager *)lm 
				   andParameters: (Parameter *)p;
-(id <ImitativeStrategy>)
     getABelowImitativeStrategyForLandManager: (LandManager *)lm 
                                andParameters: (Parameter *)p;
-(id <NonImitativeStrategy, NonHistoricalStrategy>)
     getAnInitialStrategyForLandManager: (LandManager *)lm 
                          andParameters: (Parameter *)p;
-(Interval *)getProb;
-loadFromFileNamed: (const char *)filename;
                                        // Just use objectloader for this
-saveToFileNamed: (const char *)filename;
                                  	// Again, use objectsaver -- you can
// use a template probeMap to tell it which instance variables to save.
-(void)writeHistory: (FILE *)fp;
-clickMeToGetStrategySelectorGUI;
-(void)displayProbe;
-(void)drop;
-(unsigned)getManagerCount;

@end
